from drf_actions.views import EventJournalViewSet
from rest_framework.routers import DefaultRouter
from django.urls import include, path


router = DefaultRouter()
router.register(r"events", EventJournalViewSet, basename="event")


urlpatterns = [path("actions/", include(router.urls))]



from django.conf import settings
from model_utils import Choices


if hasattr(settings, "DRF_ACTIONS_SETTINGS"):
    DRF_ACTIONS_SETTINGS = settings.DRF_ACTIONS_SETTINGS
else:
    DRF_ACTIONS_SETTINGS = {
        "content_types": {
            "user": {
                "pk": "id",
                "owner": None,
                "catch_update": [],
                "m2m": [
                    (
                        "users_user_groups",
                        "group_id",
                        "user_id",
                        "auth_group",
                        "id",
                        "name",
                        "groups",
                    )
                ],
                "model": ("users", "user"),
                "fields": (
                    ("email", "email"),
                    ("photo", "photo"),
                    ("partner", "partner_id"),
                    ("full_name", "full_name",),
                ),
            }
        },
        "create_event_for_new_entity": True,
        "queue": "events",
        "route": "main",
    }


ACTION_EVENTS = DRF_ACTIONS_SETTINGS.get(
    "events", Choices("INSERT", "UPDATE", "DELETE",)
)

ACTION_CONTENT_TYPES = Choices(*DRF_ACTIONS_SETTINGS["content_types"].keys())

INIT_CONTENT_TYPE = DRF_ACTIONS_SETTINGS.get("create_event_for_new_entity", True)

ACTION_QUEUE = DRF_ACTIONS_SETTINGS.get("queue", "events")

ACTION_ROUTE = DRF_ACTIONS_SETTINGS.get("route", "main")



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from drf_actions.models import ActionContentType
from drf_actions.app_settings import INIT_CONTENT_TYPE


@receiver(post_delete, sender=ActionContentType)
def drop_trigger_signal(sender, instance, using, **kwargs):
    instance.drop_trigger()


@receiver(post_save, sender=ActionContentType)
def create_trigger_signal(sender, instance, created, **kwargs):
    if created:
        instance.init_triggers()

        if INIT_CONTENT_TYPE:
            instance.create_current_type_events()



# Generated by Django 2.2.13 on 2020-06-18 10:44

from django.db import migrations
from django.db import connection
from drf_actions.app_settings import ACTION_QUEUE, ACTION_ROUTE


def init_schema():
    query = "CREATE SCHEMA IF NOT EXISTS rabbitmq;"

    with connection.cursor() as cursor:
        cursor.execute(query)


def init_send_message():
    query = """create or replace function rabbitmq.send_message(channel text, routing_key text, message text)
            returns void as $$ select	pg_notify(channel, routing_key || '|' || message); $$ stable language sql;"""

    with connection.cursor() as cursor:
        cursor.execute(query)


def init_on_row_change():
    query = f"""create or replace function rabbitmq.on_row_change() returns trigger as $$
      declare
        routing_key text;
        row record;
      begin
        routing_key := '{ACTION_ROUTE}';
        if (TG_OP = 'DELETE') then
            row := old;
        elsif (TG_OP = 'UPDATE') then
            row := new;
        elsif (TG_OP = 'INSERT') then
            row := new;
        end if;
        -- change 'events' to the desired channel/exchange name
        perform rabbitmq.send_message('{ACTION_QUEUE}', routing_key, row_to_json(row)::text);
        return null;
      end;
    $$ stable language plpgsql;"""

    with connection.cursor() as cursor:
        cursor.execute(query)


def init_trigger():
    drop_query = (
        f"DROP TRIGGER IF EXISTS send_change_event ON drf_actions_eventjournal;"
    )
    query = """create trigger send_change_event after insert or update or delete on drf_actions_eventjournal 
    for each row execute procedure rabbitmq.on_row_change();"""
    with connection.cursor() as cursor:
        cursor.execute(drop_query)
        cursor.execute(query)


def init_listen_events(apps, schema_editor):
    init_schema()
    init_send_message()
    init_on_row_change()
    init_trigger()


class Migration(migrations.Migration):
    dependencies = [
        ("drf_actions", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(init_listen_events),
    ]



from django.contrib import admin
from django.db import models
from django_json_widget.widgets import JSONEditorWidget
from drf_actions.models import ActionContentType, EventJournal


@admin.register(ActionContentType)
class ActionContentTypeAdmin(admin.ModelAdmin):
    list_display = ["id", "content_type", "table"]


@admin.register(EventJournal)
class EventJournalAdmin(admin.ModelAdmin):
    list_display = ["id", "reason", "object_id", "content_type"]

    formfield_overrides = {
        models.JSONField: {"widget": JSONEditorWidget},
    }



from rest_framework import viewsets, pagination, filters
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework_api_key.permissions import HasAPIKey
from drf_actions.serializers.event_journal import EventJournalSerializer
from drf_actions.models import EventJournal


class StandardPagination(pagination.PageNumberPagination):
    page_size = 100
    page_size_query_param = "page_size"
    max_page_size = 10000


class EventJournalViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = [
        filters.SearchFilter,
        DjangoFilterBackend,
        filters.OrderingFilter,
    ]
    serializer_class = EventJournalSerializer
    queryset = EventJournal.objects.all().order_by("-created")
    permission_classes = (HasAPIKey,)
    pagination_class = StandardPagination
    filter_fields = {
        "id": ["exact", "in", "lt", "lte", "gt", "gte"],
        "created": ["range", "lt", "gt", "lte", "gte"],
        "reason": ["exact", "in"],
        "content_type": ["exact", "in"],
    }
    ordering_fields = (
        "id",
        "created",
    )



# Generated by Django 3.2.5 on 2021-07-29 07:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('drf_actions', '0002_auto_20200618_1044'),
    ]

    operations = [
        migrations.AlterField(
            model_name='actioncontenttype',
            name='content_type',
            field=models.CharField(choices=[('user', 'user'), ('book', 'book'), ('publishbook', 'publishbook')], db_index=True, max_length=50),
        ),
        migrations.AlterField(
            model_name='actioncontenttype',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='eventjournal',
            name='content_type',
            field=models.CharField(choices=[('user', 'user'), ('book', 'book'), ('publishbook', 'publishbook')], db_index=True, max_length=50),
        ),
        migrations.AlterField(
            model_name='eventjournal',
            name='data',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='eventjournal',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='eventjournal',
            name='reason',
            field=models.CharField(choices=[('INSERT', 'INSERT'), ('UPDATE', 'UPDATE'), ('DELETE', 'DELETE')], db_index=True, max_length=30),
        ),
    ]



from rest_framework import serializers
from drf_actions.models import EventJournal


class EventJournalSerializer(serializers.ModelSerializer):
    class Meta:
        model = EventJournal
        fields = (
            "id",
            "reason",
            "object_id",
            "content_type",
            "data",
            "created",
            "modified",
        )



# Generated by Django 2.2.13 on 2020-06-18 10:39

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ActionContentType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "content_type",
                    models.CharField(
                        choices=[("book", "book"), ("user", "user")],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                ("table", models.CharField(max_length=100)),
            ],
            options={"abstract": False,},
        ),
        migrations.CreateModel(
            name="EventJournal",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "reason",
                    models.CharField(
                        choices=[
                            ("create", "create"),
                            ("update", "update"),
                            ("delete", "delete"),
                        ],
                        db_index=True,
                        max_length=30,
                    ),
                ),
                ("object_id", models.CharField(db_index=True, max_length=100)),
                (
                    "content_type",
                    models.CharField(
                        choices=[("book", "book"), ("user", "user")],
                        db_index=True,
                        max_length=50,
                    ),
                ),
                (
                    "data",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        blank=True, null=True
                    ),
                ),
            ],
            options={"abstract": False,},
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['drf_actions',
 'drf_actions.migrations',
 'drf_actions.serializers',
 'drf_actions.views']

package_data = \
{'': ['*']}

install_requires = \
['Django>3.1.6',
 'django-filter>=2.4.0',
 'django-model-utils>=4.1.1',
 'djangorestframework>=3.12.2']

setup_kwargs = {
    'name': 'drf-actions',
    'version': '0.2.1',
    'description': 'Create event log with help triggers and send notify after create event',
    'long_description': '# drf-actions\n\nCreate event log with help triggers and send notify after create event\n',
    'author': 'Pavel Maltsev',
    'author_email': 'pavel@speechki.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/speechki-book/drf-actions',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.db import models
from django.db.models.fields import files
from django.apps import apps
from model_utils.models import TimeStampedModel
from typing import Type, Tuple, List, Dict
from drf_actions.app_settings import (
    DRF_ACTIONS_SETTINGS,
    ACTION_EVENTS,
    ACTION_CONTENT_TYPES,
    INIT_CONTENT_TYPE,
)
from itertools import islice
from django.db import connection


class EventJournal(TimeStampedModel):
    REASONS = ACTION_EVENTS
    CONTENT_TYPES = ACTION_CONTENT_TYPES

    reason = models.CharField(max_length=30, choices=REASONS, db_index=True)
    object_id = models.CharField(max_length=100, db_index=True)
    content_type = models.CharField(max_length=50, choices=CONTENT_TYPES, db_index=True)
    data = models.JSONField(blank=True, null=True)

    def __str__(self):
        return f"{self.content_type} : {self.reason} : {self.object_id}"


class ActionContentType(TimeStampedModel):
    CONTENT_TYPES = ACTION_CONTENT_TYPES

    content_type = models.CharField(max_length=50, choices=CONTENT_TYPES, db_index=True)
    table = models.CharField(max_length=100)

    @staticmethod
    def prepare_event_journal(
        objs,
        content_type: str,
        pk: str,
        fields_dict: Dict[str, str],
        m2m: List[Tuple[str, str]],
    ):
        owner = DRF_ACTIONS_SETTINGS["content_types"][content_type].get("owner")
        for obj in objs:
            data = {}
            for key, value in fields_dict.items():
                attr_value = getattr(obj, key)
                if isinstance(attr_value, files.FieldFile) and attr_value.name:
                    data[value] = attr_value.url
                elif isinstance(attr_value, files.FieldFile) and not attr_value.name:
                    data[value] = None
                else:
                    data[value] = attr_value

            for m2m_field, attr_name in m2m:
                data[m2m_field] = [
                    getattr(item, attr_name) for item in getattr(obj, m2m_field).all()
                ]

            if owner:
                data["owner"] = getattr(obj, owner)

            yield EventJournal(
                reason=ACTION_EVENTS.INSERT,
                content_type=content_type,
                object_id=getattr(obj, pk),
                data=data,
            )

    def bulk_create_entities(self, model: Type["models.Model"], content_type: str):
        m2m = [
            (item[6], item[5])
            for item in DRF_ACTIONS_SETTINGS["content_types"][content_type]["m2m"]
        ]
        prefetch = [item[0] for item in m2m]
        pk = DRF_ACTIONS_SETTINGS["content_types"][content_type]["pk"]
        fields_dict = {
            item[1]: item[0]
            for item in DRF_ACTIONS_SETTINGS["content_types"][content_type]["fields"]
        }
        fields_dict[pk] = pk
        current_objects = set(
            EventJournal.objects.filter(content_type=content_type).values_list(
                pk, flat=True
            )
        )
        queryset = set(model.objects.all().values_list(pk, flat=True))
        obj_ids = queryset.difference(current_objects)
        objs = model.objects.prefetch_related(*prefetch).filter(pk__in=obj_ids)

        new_objs = self.prepare_event_journal(objs, content_type, pk, fields_dict, m2m)

        while True:
            batch = list(islice(new_objs, 100))
            if not batch:
                break

            EventJournal.objects.bulk_create(batch, 100)

    def create_current_type_events(self):
        app_name, model_name = DRF_ACTIONS_SETTINGS["content_types"][self.content_type][
            "model"
        ]
        model: Type["models.Model"] = apps.get_model(
            app_label=app_name, model_name=model_name
        )
        self.bulk_create_entities(model, self.content_type)

    @classmethod
    def create_events(cls):
        content_types = cls.objects.values_list("content_type", flat=True)
        for key, content_type in DRF_ACTIONS_SETTINGS["content_types"].values():
            if key not in content_types:
                app_name, model_name = content_type["model"]
                model: Type["models.Model"] = apps.get_model(
                    app_label=app_name, model_name=model_name
                )
                cls.objects.create(content_type=key, table=model._meta.db_table)

    @classmethod
    def delete_content_types(cls, content_types):
        for_delete_content_types = cls.objects.exclude(content_type__in=content_types)
        for content_type in for_delete_content_types:
            content_type.delete()

    @classmethod
    def init_content_type(cls):
        if INIT_CONTENT_TYPE:
            cls.create_events()

    @staticmethod
    def trigger_prefix():
        return "action_"

    def drop_trigger(self):
        prefix = self.trigger_prefix()
        query = f"DROP TRIGGER IF EXISTS {prefix}{self.content_type}_trigger ON {self.table};"
        func = f"DROP FUNCTION IF EXISTS public.{prefix}{self.content_type}_func"

        with connection.cursor() as cursor:
            cursor.execute(query)
            cursor.execute(func)

    def m2m_init(self):
        m2m = DRF_ACTIONS_SETTINGS["content_types"][self.content_type].get("m2m")

        if not m2m:
            return "", "", []

        m2m_init_variables = []
        m2m_calc_variables = []
        m2m_values = []
        for item in m2m:
            m2m_init_variables.append(f"{item[6]} json;")
            m2m_calc_variables.append(
                f"{item[6]}:= json_agg(tmp.{item[5]})::jsonb from (select name from {item[0]} left join {item[3]} ag on"
                f" {item[0]}.{item[1]} = ag.{item[4]} where {item[0]}.{item[2]} = NEW.id) as tmp;"
            )
            m2m_values.append(f"'{item[6]}',{item[6]}")

        return "\n".join(m2m_init_variables), "\n".join(m2m_calc_variables), m2m_values

    def owner(self):
        owner = DRF_ACTIONS_SETTINGS["content_types"][self.content_type].get("owner")

        if not owner:
            return []

        return ["'owner'", f"NEW.{owner}"]

    def create_trigger_function(self):
        prefix = self.trigger_prefix()
        build_json = []
        delete_build_json = []
        for field in DRF_ACTIONS_SETTINGS["content_types"][self.content_type]["fields"]:
            build_json.append(f"'{field[0]}'")
            build_json.append(f"NEW.{field[1]}")

            delete_build_json.append(f"'{field[0]}'")
            delete_build_json.append(f"OLD.{field[1]}")

        m2to_variables, m2m_calc, m2m_values = self.m2m_init()
        owner = self.owner()

        build_json_str = ",".join(build_json + m2m_values + owner)
        delete_build_json_str = ",".join(delete_build_json + owner)
        pk = DRF_ACTIONS_SETTINGS["content_types"][self.content_type]["pk"]

        func = f"""CREATE OR REPLACE FUNCTION public.{prefix}{self.content_type}_func() RETURNS TRIGGER AS
                $BODY$
                declare
                    json_new jsonb;
                    reason varchar;
                    {m2to_variables}
                BEGIN
                    IF tg_op = 'DELETE' THEN
                        json_new:= json_build_object({delete_build_json_str});
                        INSERT INTO {EventJournal._meta.db_table}(reason, object_id, content_type, data, created, modified)
                            VALUES (tg_op ,OLD.{pk}, '{self.content_type}', json_new, now(), now());
                    ELSE
                        {m2m_calc}
                        json_new:= json_build_object({build_json_str});
                        INSERT INTO {EventJournal._meta.db_table}(reason, object_id, content_type, data, created, modified)
                            VALUES (tg_op ,NEW.{pk}, '{self.content_type}', json_new, now(), now());
                    END IF;
                        RETURN null; 
                END; $BODY$ LANGUAGE plpgsql;"""

        with connection.cursor() as cursor:
            cursor.execute(func)

    def init_triggers(self):
        prefix = self.trigger_prefix()

        catch_update = DRF_ACTIONS_SETTINGS["content_types"][self.content_type].get(
            "catch_update"
        )

        if catch_update:
            catch_fields = " of " + ",".join(catch_update)
        else:
            catch_fields = ""

        trigger = f"""create trigger {prefix}{self.content_type}_trigger
                                after insert or update {catch_fields} or delete on {self.table}
                                for each row execute procedure public.{prefix}{self.content_type}_func();
                            """

        self.create_trigger_function()
        with connection.cursor() as cursor:
            cursor.execute(trigger)
